#include "Zeolite.h"
#include "zMesh.h"
#include "vectors.h"
#include <zbuffer.h>
#include <zList.h>

CzMesh::CzMesh(void) {
	m_hName = NULL;
	m_hAxes = NULL;
	m_hOffset = NULL;
	m_hScale = NULL;
	m_hVertBuf = NULL;
	m_hNormBuf = NULL;
	m_hTexBuf = NULL;
	m_hFaceGroupList = NULL;
	m_hMtlList = NULL;
	m_hSubMeshList = NULL;
	m_hInfoEx = NULL;
}

CzMesh::~CzMesh(void) {
	
}

bool CzMesh::Create() {
	
	if(m_hVar) {
		zeoReportError("CzMesh::Create error:\r\n - variable is already created");
		return false;
	}

	if(!CzVar::Create(VarID_varlist)) {
		zeoReportError("CzMesh::Create error:\r\n - call to CzVar::Create failed");
		return false;
	}

	if(!CreateData()) {
		zeoReportError("CzMesh::Create error:\r\n - call to CzMesh::CreateData failed");
		return false;
	}

	return true;
}
bool CzMesh::CreateShared(const char* lpVarName) {
	
	if(m_hVar) {
		zeoReportError("CzMesh::CreateShared error:\r\n - variable is already created");
		return false;
	}

	if(!CzVar::CreateShared(VarID_varlist, lpVarName)) {
		zeoReportError("CzMesh::CreateShared error:\r\n - call to CzVar::CreateShared failed");
		return false;
	}

	if(!CreateData()) {
		zeoReportError("CzMesh::CreateShared error:\r\n - call to CzMesh::CreateData failed");
		return false;
	}

	return true;
}

bool CzMesh::CreateData(ZLIST hMeshTarget) {

	if(hMeshTarget) {
		if(!CzVar::Attach(hMeshTarget)) {
			zeoReportError("CzMesh::CreateData error:\r\n - call to CzVar::Attach failed for target");
			return false;
		}
	}

	if(!m_hVar) {
		zeoReportError("CzMesh::CreateData error:\r\n - mesh variable not created");
		return false;
	}

	if(zlist_nItems(m_hVar)) {
		zeoReportError("CzMesh::CreateData error:\r\n - mesh not initialised");
		return false;
	}

	// root level items follow
	int MeshVer = 1;
	if(!zvar_SetValue(zlist_CreateItem(m_hVar, VarID_int, "MeshVer"), &MeshVer)) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot set mesh version");
		return false;
	}

	if(! (m_hName = zlist_CreateItem(m_hVar, VarID_string, "Name"))) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create mesh name");
		return false;
	}

	matrix3f axes = MatrixIdentity3f();
	if(!zvar_SetValue( ( m_hAxes = zlist_CreateItem(m_hVar, VarID_matrix3f, "Axes") ), &axes)) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create mesh axes");
		return false;
	}

	vector3f vo = VectorSet3f(0, 0, 0);
	if(!zvar_SetValue( ( m_hOffset = zlist_CreateItem(m_hVar, VarID_vector3f, "Offset") ), &vo)) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create mesh offset");
		return false;
	}

	vector3f vs = VectorSet3f(1, 1, 1);
	if(!zvar_SetValue( ( m_hScale = zlist_CreateItem(m_hVar, VarID_vector3f, "Scale") ), &vs)) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create mesh scale");
		return false;
	}

	if(! (m_hVertBuf = zlist_CreateItem(m_hVar, VarID_buffer, "VertBuf"))) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create mesh vertex buffer");
		return false;
	}

	if(! (m_hNormBuf = zlist_CreateItem(m_hVar, VarID_buffer, "NormBuf"))) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create mesh normal buffer");
		return false;
	}

	if(! (m_hTexBuf = zlist_CreateItem(m_hVar, VarID_buffer, "TexBuf"))) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create mesh texture buffer");
		return false;
	}

	if(! (m_hFaceGroupList = zlist_CreateItem(m_hVar, VarID_varlist, "FaceGroupList"))) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create face group list");
		return false;
	}

	if(! (m_hMtlList = zlist_CreateItem(m_hVar, VarID_varlist, "MtlLib"))) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create material library");
		return false;
	}

	if(! (m_hSubMeshList = zlist_CreateItem(m_hVar, VarID_varlist, "SubMeshList"))) {
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create sub mesh list");
		return false;
	}

	if(! (m_hInfoEx = zlist_CreateItem(m_hVar, VarID_varlist, "InfoEx"))) { // extra data not handled by base mesh, but used by renderer or file format
		zeoReportError("CzMesh::CreateData error:\r\n - cannot create InfoEx");
		return false;
	}

	return true;
}

bool CzMesh::Attach(ZVAR hVar) {
	
	if(!CzVar::Attach(hVar)) {
		zeoReportError("CzMesh::Attach error:\r\n - call to CzVar::Attach failed");
		return false;
	}

	if(!LoadHandles()) {
		zeoReportError("CzMesh::Attach error:\r\n - call to CzMesh::LoadHandles failed");
		return false;
	}

	return true;
}

bool CzMesh::LoadHandles() {

	int argid = 0;

	// get mesh version
	int MeshVer = -1;
	if(!zvar_GetValueEx(zlist_GetItemI(m_hVar, argid++), VarID_int, &MeshVer)) {
		zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get mesh version");

		Release();
		return false;
	}

	switch(MeshVer) {
	case -1:
		zeoReportError("CzMesh::LoadHandles error:\r\n - invalid mesh version");

		Release();
		return false;
	default:
		zeoReportError("CzMesh::LoadHandles error:\r\n - unsupported mesh version");

		Release();
		return false;
	case 1:
		{
			if(!zvar_IsType(m_hName = zlist_GetItemI(m_hVar, argid++), VarID_string)) { // name at 1
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get mesh axes");
				return false;
			}

			if(!zvar_IsType(m_hAxes = zlist_GetItemI(m_hVar, argid++), VarID_matrix3f)) { // axes at 2
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get mesh axes");
				return false;
			}
		
			if(!zvar_IsType (m_hOffset = zlist_GetItemI(m_hVar, argid++), VarID_vector3f)) { // offset at 3
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get mesh offset");
				return false;
			}

			if(!zvar_IsType (m_hScale = zlist_GetItemI(m_hVar, argid++), VarID_vector3f)) { // scale at 4
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get mesh scale");
				return false;
			}

			if(!zvar_IsType (m_hVertBuf = zlist_GetItemI(m_hVar, argid++), VarID_buffer)) { // vertex buffer at 5
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get mesh vertex buffer");
				return false;
			}

			if(!zvar_IsType (m_hNormBuf = zlist_GetItemI(m_hVar, argid++), VarID_buffer)) { // normal buffer at 6
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get mesh normal buffer");
				return false;
			}

			if(!zvar_IsType (m_hTexBuf = zlist_GetItemI(m_hVar, argid++), VarID_buffer)) { // texture buffer at 8
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get mesh texture buffer");
				return false;
			}

			if(!zvar_IsType (m_hFaceGroupList = zlist_GetItemI(m_hVar, argid++), VarID_varlist)) { // face group list at 9
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get face group list");
				return false;
			}

			if(!zvar_IsType (m_hMtlList = zlist_GetItemI(m_hVar, argid++), VarID_varlist)) { // material library at 10
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get material library");
				return false;
			}

			if(!zvar_IsType (m_hSubMeshList = zlist_GetItemI(m_hVar, argid++), VarID_varlist)) { // sub mesh list at 11
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get sub mesh list");
				return false;
			}

			if(!zvar_IsType (m_hInfoEx = zlist_GetItemI(m_hVar, argid++), VarID_varlist)) { // infoex at 12
				zeoReportError("CzMesh::LoadHandles error:\r\n - cannot get InfoEx");
				return false;
			}
		}
		break;
	}

	return true;
}

const char* CzMesh::GetMeshName() { return zstr_GetText(m_hName); }
bool CzMesh::SetMeshName(const char* lpMeshName) {
	return zstr_SetText(m_hName, lpMeshName);
}

bool CzMesh::GetAxes(matrix3f** ppAxes) {
	if(!ppAxes)	return false;
	matrix3f* pValue = (matrix3f*)zvar_GetDataPtr(m_hAxes);
	if(!pValue) return false;

	*ppAxes = pValue;
	return true;
}
bool CzMesh::GetOffset(vector3f** ppOffset) {
	if(!ppOffset)	return false;
	vector3f* pValue = (vector3f*)zvar_GetDataPtr(m_hOffset);
	if(!pValue) return false;

	*ppOffset = pValue;
	return true;
}
bool CzMesh::GetScale(vector3f** ppScale) {
	if(!ppScale)	return false;
	vector3f* pValue = (vector3f*)zvar_GetDataPtr(m_hScale);
	if(!pValue) return false;

	*ppScale = pValue;
	return true;
}

bool CzMesh::SetCoords(ZBUF hVertBuf, ZBUF hNormBuf, ZBUF hTexBuf) {
	if(!m_hVar)	return false;

	if(VarID_vector3f!=zbuffer_ItemType(hVertBuf)) {
		zeoReportError("CzMesh::SetCoords error:\r\n - invalid vertex buffer type (vector3f expected)");
		return false;
	}
	if(!zvar_CopyValue(GetVertBuf(), hVertBuf)) {
		zeoReportError("CzMesh::SetCoords error:\r\n - cannot copy vert buffer");
		return false;
	}

	if(hNormBuf) {
		if(zbuffer_nItems(hNormBuf)) {
			if(VarID_vector3f!=zbuffer_ItemType(hNormBuf)) {
				zeoReportError("CzMesh::SetCoords error:\r\n - invalid norm buffer type (vector3f expected)");
				return false;
			}
			if(!zvar_CopyValue(GetNormBuf(), hNormBuf)) {
				zeoReportError("CzMesh::SetCoords error:\r\n - cannot copy norm buffer");
				return false;
			}
		}
	}
	if(hTexBuf) {
		if(zbuffer_nItems(hTexBuf)) {
			if(VarID_vector3f!=zbuffer_ItemType(hTexBuf)) {
				zeoReportError("CzMesh::SetCoords error:\r\n - invalid tex buffer type (vector3f expected)");
				return false;
			}
			if(!zvar_CopyValue(GetTexBuf(), hTexBuf)) {
				zeoReportError("CzMesh::SetCoords error:\r\n - cannot copy tex buffer");
				return false;
			}
		}
	}

	return true;
}


long CzMesh::nFaceGroups() {
	if(!m_hVar)	return 0;

	return zlist_nItems(m_hFaceGroupList);
}
long CzMesh::NewFaceGroup(CzMeshFaceGroup &mfg, const char* lpFaceGroupName) {
	if(!m_hVar)	return -1;

	ZLIST hItem = zlist_CreateItem(m_hFaceGroupList, VarID_varlist, lpFaceGroupName);
	if(!hItem) {
		zeoReportError("CzMesh::NewFaceGroup error:\r\n - cannot create new item");
		return -1;
	}

	long rval = zlist_nItems(m_hFaceGroupList)-1;
	if(!mfg.CreateData(hItem)) {
		zeoReportError("CzMesh::NewFaceGroup error:\r\n - call to CzMeshFaceGroup::CreateData failed");
		zlist_DelItemI(m_hFaceGroupList, rval);
		return -1;
	}

	return rval;
}
bool CzMesh::GetFaceGroup(long index, CzMeshFaceGroup &mfg) {
	if(!m_hVar)	return false;

	ZLIST hItem = zlist_GetItemI(m_hFaceGroupList, index);
	if(!hItem) {
		zeoReportError("CzMesh::GetFaceGroup error:\r\n - cannot get item");
		return false;
	}

	if(!mfg.Attach(hItem)) {
		zeoReportError("CzMesh::GetFaceGroup error:\r\n - call to CzMeshFaceGroup::Attach failed");
		return false;
	}

	return true;
}
long CzMesh::FindFaceGroup(const char* lpFaceGroupName, long StartAt) {
	if(!m_hVar)	return -1;

	long nItems = nFaceGroups();
	for(int i=StartAt; i<nItems; i++) {
		ZVAR hVar = zlist_GetItemI(m_hFaceGroupList, i);
		if(!hVar) {
			zeoReportError("CzMesh::FindFaceGroup error:\r\n - call to cannot get sub item handle");
			return -1;
		}
		if(0==strcmp(lpFaceGroupName, zvar_GetName(hVar))) {
			return i;
		}
	}

	return -1;
}
bool CzMesh::DeleteFaceGroup(long index) {
	if(!m_hVar)	return false;

	if(!zlist_DelItemI(m_hFaceGroupList, index)) {
		zeoReportError("CzMesh::DeleteFaceGroup error:\r\n - delete failed");
		return false;
	}

	return true;
}

long CzMesh::nMaterials() {
	if(!m_hVar)	return 0;

	return zlist_nItems(m_hMtlList);
}
long CzMesh::NewMaterial(CzMeshMtl &mtl, const char* lpMtlName) {

	if(!m_hVar)	return -1;

	ZLIST hItem = zlist_CreateItem(m_hMtlList, VarID_varlist, lpMtlName);
	if(!hItem) {
		zeoReportError("CzMesh::NewMaterial error:\r\n - cannot create new item");
		return -1;
	}

	long rval = zlist_nItems(m_hMtlList)-1;

	if(!mtl.CreateData(hItem)) {
		zeoReportError("CzMesh::NewMaterial error:\r\n - call to CzMeshMtl::CreateData failed");
		zlist_DelItemI(m_hMtlList, rval);
		return -1;
	}

	return rval;
}
bool CzMesh::GetMaterial(long index, CzMeshMtl &mtl) {
	if(!m_hVar)	return false;

	ZLIST hItem = zlist_GetItemI(m_hMtlList, index);
	if(!hItem) {
		zeoReportError("CzMesh::GetMaterial error:\r\n - cannot get item");
		return false;
	}

	if(!mtl.Attach(hItem)) {
		zeoReportError("CzMesh::GetMaterial error:\r\n - call to CzMeshMtl::Attach failed");
		return false;
	}

	return true;
}
long CzMesh::FindMaterial(const char* lpMtlName, long StartAt) {
	if(!m_hVar)	return -1;

	long nItems = nMaterials();
	for(int i=StartAt; i<nItems; i++) {
		ZVAR hVar = zlist_GetItemI(m_hMtlList, i);
		if(!hVar) {
			zeoReportError("CzMesh::FindMaterial error:\r\n - call to cannot get sub item handle");
			return -1;
		}
		if(0==strcmp(lpMtlName, zvar_GetName(hVar))) {
			return i;
		}
	}

	return -1;
}
bool CzMesh::DeleteMaterial(long index) {
	if(!m_hVar)	return false;

	if(!zlist_DelItemI(m_hMtlList, index)) {
		zeoReportError("CzMesh::DeleteMaterial error:\r\n - delete failed");
		return false;
	}

	return true;
}

long CzMesh::nSubMeshes() {
	if(!m_hVar)	return 0;

	return zlist_nItems(m_hSubMeshList);
}
long CzMesh::NewSubMesh(CzMesh &mesh, const char* lpSubMeshName) {

	if(!m_hVar)	return -1;

	ZLIST hItem = zlist_CreateItem(m_hSubMeshList, VarID_varlist, NULL);
	if(!hItem) {
		zeoReportError("CzMesh::NewSubMesh error:\r\n - cannot create new item");
		return -1;
	}

	long rval = zlist_nItems(m_hSubMeshList)-1;

	if(!mesh.CreateData(hItem)) {
		zeoReportError("CzMesh::NewSubMesh error:\r\n - call to CzMesh::CreateData failed");
		zlist_DelItemI(m_hSubMeshList, rval);
		return -1;
	}

	if(lpSubMeshName) {
		if(!mesh.SetMeshName(lpSubMeshName)) {
			zeoReportError("CzMesh::NewSubMesh error:\r\n - cannot set mesh name");
			zlist_DelItemI(m_hSubMeshList, rval);
			return -1;
		}
	}

	return rval;
}
bool CzMesh::GetSubMesh(long index, CzMesh &mesh) {
	if(!m_hVar)	return false;

	ZLIST hItem = zlist_GetItemI(m_hSubMeshList, index);
	if(!hItem) {
		zeoReportError("CzMesh::GetSubMesh error:\r\n - cannot get item");
		return false;
	}

	if(!mesh.Attach(hItem)) {
		zeoReportError("CzMesh::GetSubMesh error:\r\n - call to CzMesh::Attach failed");
		return false;
	}

	return true;
}
long CzMesh::FindSubMesh(const char* lpSubMeshName, long StartAt) {
	if(!m_hVar)	return -1;

	long nItems = nSubMeshes();
	for(int i=StartAt; i<nItems; i++) {
		CzMesh m;
		if(!GetSubMesh(i, m)) {
			zeoReportError("CzMesh::FindSubMesh error:\r\n - cannot get sub mesh");
			return -1;
		}
		if(0==strcmp(lpSubMeshName, m.GetMeshName())) {
			return i;
		}
	}

	return -1;
}
bool CzMesh::DeleteSubMesh(long index) {
	if(!m_hVar)	return false;

	if(!zlist_DelItemI(m_hSubMeshList, index)) {
		zeoReportError("CzMesh::DeleteSubMesh error:\r\n - delete failed");
		return false;
	}

	return true;
}

bool CzMesh::Validate() {

	if(!m_hVar) return false;

	CzBuffer VertBuf;
	if(!VertBuf.Attach(m_hVertBuf)) {
		zeoWriteToLog("CzMesh::Validate error:\r\n - cannot attach vertex buffer");
		return false;
	}
	if(VertBuf.nItems()) {
		if(VarID_vector3f!=VertBuf.ItemType()) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - vertex buffer is not type vector3f");
			return false;
		}
	}

	CzBuffer NormBuf;
	if(!NormBuf.Attach(m_hNormBuf)) {
		zeoWriteToLog("CzMesh::Validate error:\r\n - cannot attach normal buffer");
		return false;
	}
	if(NormBuf.nItems()) {
		if(VarID_vector3f!=NormBuf.ItemType()) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - normal buffer is not type vector3f");
			return false;
		}
	}

	CzBuffer TexBuf;
	if(!TexBuf.Attach(m_hTexBuf)) {
		zeoWriteToLog("CzMesh::Validate error:\r\n - cannot attach texture buffer");
		return false;
	}
	if(TexBuf.nItems()) {
		if(VarID_vector3f!=TexBuf.ItemType()) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - texture buffer is not type vector3f");
			return false;
		}
	}


	// iterate through face groups to check their indices
	for(int f=0; f<nFaceGroups(); f++) {
		CzMeshFaceGroup fg;
		if(!GetFaceGroup(f, fg)) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get face group");
			return false;
		}

		CzBuffer VertIndexBuf;
		if(!VertIndexBuf.Attach(fg.GetVertIndexBuf())) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get face group vert index buffer");
			return false;
		}
		if(VarID_int!=VertIndexBuf.ItemType()) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - face group vert index buffer is not initialised as type int");
			return false;
		}

		ZBUF hNormIndexBuf = fg.GetNormIndexBuf();
		if(!hNormIndexBuf) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get face group norm index buffer");
			return false;
		}
		if(0==zbuffer_nItems(hNormIndexBuf)) {
			hNormIndexBuf = NULL;
		} else {
			if(VertIndexBuf.nItems()!=zbuffer_nItems(hNormIndexBuf)) {
				zeoWriteToLog("CzMesh::Validate error:\r\n - face group norm index buffer is different size to face group vert index buffer");
				return false;
			}
			if(VarID_int!=zbuffer_ItemType(hNormIndexBuf)) {
				zeoWriteToLog("CzMesh::Validate error:\r\n - face group norm index buffer is not initialised as type int");
				return false;
			}
			if(!NormBuf.nItems()) {
				zeoWriteToLog("CzMesh::Validate error:\r\n - face group norm index buffer exists, but mesh norm buffer is empty");
				return false;
			}
		}

		ZBUF hTexIndexBuf = fg.GetTexIndexBuf();
		if(!hTexIndexBuf) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get face group tex index buffer");
			return false;
		}
		if(0==zbuffer_nItems(hTexIndexBuf)) {
			hTexIndexBuf = NULL;
		} else {
			if(VertIndexBuf.nItems()!=zbuffer_nItems(hTexIndexBuf)) {
				zeoWriteToLog("CzMesh::Validate error:\r\n - face group tex index buffer is different size to face group vert index buffer");
				return false;
			}
			if(VarID_int!=zbuffer_ItemType(hTexIndexBuf)) {
				zeoWriteToLog("CzMesh::Validate error:\r\n - face group tex index buffer is not initialised as type int");
				return false;
			}
			if(!TexBuf.nItems()) {
				zeoWriteToLog("CzMesh::Validate error:\r\n - face group tex index buffer exists, but mesh tex buffer is empty");
				return false;
			}
		}

		for(int i=0; i<VertIndexBuf.nItems(); i++) {
			int index = -1;
			if(!VertIndexBuf.GetValue(i, &index)) {
				zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get value from face group vert index buffer");
				return false;
			}
			vector3f v;
			if(!VertBuf.GetValue(index, &v)) {
				zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get value from vert buffer (invalid index in face group vert index buffer)");
				return false;
			}

			if(hNormIndexBuf) {
				if(!zbuffer_GetValue(hNormIndexBuf, i, &index)) {
					zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get value from face group norm index buffer");
					return false;
				}
				if(!NormBuf.GetValue(index, &v)) {
					zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get value from norm buffer (invalid index in face group norm index buffer)");
					return false;
				}
			}

			if(hTexIndexBuf) {
				if(!zbuffer_GetValue(hTexIndexBuf, i, &index)) {
					zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get value from face group tex index buffer");
					return false;
				}
				if(!TexBuf.GetValue(index, &v)) {
					zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get value from tex buffer (invalid index in face group tex index buffer)");
					return false;
				}
			}
		}
	}

	// validate submeshes
	for(int i=0; i<nSubMeshes(); i++) {
		CzMesh m;
		if(!GetSubMesh(i, m)) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - cannot get sub mesh");
			return false;
		}

		if(!m.Validate()) {
			zeoWriteToLog("CzMesh::Validate error:\r\n - sub mesh failed validation");
			return false;
		}
	}

	return true;
}